var crypto = require('crypto');
var exec = require('child_process').exec;
var fs = require('fs');
var path = require("path");
var request = require('request');
var url = require('url');
var zlib = require('zlib');

process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";

var proxy = require('./proxy.js');

var GR = { // State
    activeState: false,
    fileChunkSize: 16 * 1024, // file chunk size when uploading files
    key: undefined, // job instance key
    server: 'https://charityengine.services/remotejobs-dev/v2/',
    sharedFolder: 'shared/', // Shared folder location
    states: { // list of states and their statuses
        default: {
            commandLine: undefined, // command line to execute
            inputFiles: {}, // list of input files
            status: 'initial', // One of initial, downloading, downloaded, running, completed, failed
            startTime: undefined, // Start time in unixtime
            endTime: undefined, // End time in unixtime
        },
    },
    timeLimit: 0,
    startTime: false, // Script start time
};

GR.startTime = timestamp();

GR.sharedFolder = __dirname + '/';

if (fs.existsSync(GR.sharedFolder + 'output') === false) {
    fs.mkdirSync(GR.sharedFolder + 'output');
}

// Read startup state
if (fs.existsSync(GR.sharedFolder + 'state.default.json') === false) {
    shutdownVm(false, 'Missing startup state');
}
try {
    // @todo encryption
    GR.states.default = JSON.parse(
        fs.readFileSync(GR.sharedFolder + 'state.default.json')
    );
    GR.states.default.status = 'initial';
    GR.timeLimit = GR.states.default.timeLimit;
    GR.key = GR.states.default.key;
} catch (e) {
    shutdownVm(false, 'Failed to parse startup state');
}

// Set up timer
// Validate timer; needs to be specified in seconds in 1h increments
if ((typeof GR.timeLimit === 'undefined') || (GR.timeLimit <= 0) || (GR.timeLimit % 3600 !== 0)) {
    shutdownVm(false, 'Incorrect timer');
}
setTimeout(function() {
    shutdownVm(true); // Shut down VM after reserved time runs out
}, GR.timeLimit * 1000);

setTimeout(function() {
    runNextState();

    pollServer();
    setInterval(pollServer, 30000);
}, 5000);

function runNextState() {
    console.log('runNextState()');
    return new Promise(function(resolve, reject) {
        let stateName = chooseNextState();
        if (stateName === false) {
            // No pending states, pollServer() will eventually get and execute a new one.
            console.log('no inactive states');
            return resolve(false);
        }

        GR.activeState = stateName;

        runState(stateName)
            .then(function(result) {
                runNextState();
                //resolve(result);
            }
        );
    });
}

function chooseNextState() {
    if (GR.states.default.status === 'initial') {
        return 'default';
    }

    for (let name in GR.states) {
        if (GR.states[name].status === 'initial') {
            return name;
        }
    }

    return false;
}

/**
 * Runs a specific state
 * @param stateName Name of the state to execute. Initial state is "default"
 */
function runState(stateName) {
    console.log('runState(' + stateName + ')');
    return new Promise(function(resolve, reject) {
        // Read and download input files
        let promises = [];
        if (typeof GR.states[stateName].inputFiles !== 'undefined') {
            for (let i in GR.states[stateName].inputFiles) {
                promises.push(downloadFile(stateName, i));
            }
        }
        Promise
            .all(promises)
            .then(function(result) {
                return downloadsCompleted(stateName, result);
            })
            .then(function(result) {
                GR.activeState = false;
                resolve(result);
            });
    });
}

function downloadFile(stateName, i) {
    return new Promise(function(resolve, reject) {
        // @todo encryption
        let fullUrl = GR.states[stateName].inputFiles[i].url;
        let parsedUrl = url.parse(fullUrl, true);
        let filename = path.basename(parsedUrl.pathname);
        if (!filename) {
            filename = i;
        }

        // If output file already exists, version it
        while (fs.existsSync(filename)) {
            console.log('file already exists', filename);

            let parsed = path.parse(filename);
            filename = parsed.name + '_new' + parsed.ext;
        }
        fs.open(filename, 'wx', function(err, fd) {
            // If a file was created because of a race condition, retry
            if ((err) && (err.code === 'EEXIST')) {
                downloadFile(stateName, i).then(function() {
                    resolve();
                });
                return;
            }

            request.get(fullUrl)
                .on('response', function() {
                    GR.states[stateName].inputFiles[i].filename = filename;
                    GR.states[stateName].inputFiles[i].status = 'downloaded';
                    resolve();
                })
                .on('error', function(err) {
                    console.log(err);
                    GR.states[stateName].inputFiles[i].status = 'failed';
                    GR.states[stateName].inputFiles[i].error = err;
                    return reject();
                })
                .pipe(fs.createWriteStream('', {fd: fd}));
        });
    });
}

/**
 * Callback when all file downloads complete
 * @param stateName Name of the state that is being executed
 * @param status[] Array of statuses of all downloaded input files
 */
function downloadsCompleted(stateName, status) {
    return new Promise(function(resolve, reject) {
        GR.states[stateName].status = 'downloaded';
        if (GR.states[stateName].commandLine) {
            // Execute specified command line
            let commandLine = GR.states[stateName].commandLine;

            // Replace command line parameters with actual file names
            for (let i in GR.states[stateName].inputFiles) {
                commandLine.replace('%FILE' + (i + 1) + '%', GR.states[stateName].inputFiles[i].filename);
            }
            for (let i in GR.states['default'].inputFiles) {
                commandLine.replace('%STARTFILE' + (i + 1) + '%', GR.states['default'].inputFiles[i].filename);
            }

            var options = {
                maxBuffer: 1024 * 1024 * 1024, // default limit is 200KB, setting to 1GB instead
            };

            // Save computation start time
            GR.states[stateName].startTime = new Date().getTime();

            // Run the script
            GR.states[stateName].status = 'running';
            exec(commandLine, options, function callback(error, stdout, stderr) {
                GR.states[stateName].endTime = new Date().getTime();

                if (error) {
                    console.log('error: ' + error);
                    GR.states[stateName].status = 'failed';
                    return reject();
                }

                GR.states[stateName].status = 'completed';
                createSnapshot(stateName).then(function() {
                    console.log('created snapshot', stateName);
                    if (GR.states[stateName].cleanupFiles) {
                        fs.readdir(GR.sharedFolder + 'output', (err, files) => {
                            for (const file of files) {
                                console.log('removing', file);
                                fs.unlinkSync(GR.sharedFolder + 'output/' + file);
                            }
                            return resolve();
                        });
                        return; // Only resolve after output file cleanup
                    }
                    return resolve();
                });
            });
            return;
        }

        // No command line
        return resolve();
    });
}

function updateRemoteState(stateName) {
    return new Promise(function(resolve, reject) {
        console.log('Updating remote state', stateName);

        let payload = {
            key: GR.key,
            stateName: stateName,
            state: GR.states[stateName],
        };
        request
            .post({
                url: GR.server + 'update-state',
                timeout: 45000,
                body: JSON.stringify(payload),
                gzip: true,
            }, function(err, response, body) {
                console.log('update-state response', body);
                if (err) {
                    return reject(err);
                }
                let output = JSON.parse(body);
                if (output.success === false) {
                    return reject(output.error);
                }
                if (output.exists === false) {
                    return resolve(false);
                }
                return resolve(output.url);
            })
        ;
    });
}

function createSnapshot(stateName) {
    return new Promise(function(resolve, reject) {
        // Create new snapshot
        GR.states[stateName].outputFiles = {};

        fs.readdir(GR.sharedFolder + 'output', function(err, files) {
            // Iterate output directory
            let promises = [];
            for (let i in files) {
                let filename = files[i];
                // Upload each file
                promises.push(uploadFile(filename));
            }
            Promise
                .all(promises)
                .then(function(status) {
                    // Merge result array into a single object. Filenames are
                    // guaranteed to be unique, so no data is lost.
                    GR.states[stateName].outputFiles = status.reduce(function(acc, x) {
                        for (var key in x) {
                            acc[key] = x[key];
                        }
                        return acc;
                    }, {});
                    // All uploads are done
                    updateRemoteState(stateName)
                        .then(function(status) {
                            console.log('state updated', JSON.stringify(GR.states[stateName], null, 4));
                            return resolve();
                        })
                    ;
                })
            ;
        });
    });
}

function uploadFile(file) {
    return new Promise(function(resolve, reject) {
        console.log('Uploading', file);
        fs.stat(GR.sharedFolder + 'output/' + file, function(error, stat) {
            if (error) {
                return reject(error);
            }

            if(stat.isFile() === false) { // Directory
                return resolve(true); // not throwing, simply ignoring
            }

            // Read file in chunks
            let size = stat.size;
            let processed = 0;
            let promises = [];
            let i = 0;
            while (processed < size) {
                let uploadSize = GR.fileChunkSize;
                if (size - processed < GR.fileChunkSize) {
                    uploadSize = size - processed;
                }
                promises.push(uploadFilePart(file, i, processed, uploadSize));
                processed += uploadSize;
                i++;
            }
            Promise
                .all(promises)
                .then(function(status) {
                    // minimize file part information
                    let fileparts = [];
                    for (let part in status) {
                        fileparts[status[part].partNumber] = status[part];
                        delete fileparts[status[part].partNumber].partNumber;
                    }
                    console.log('all parts uploaded', fileparts);
                    // All parts have been uploaded
                    let result = {};
                    result[file] = {
                        size: size,
                        parts: fileparts,
                    };
                    return resolve(result);
                })
            ;
        });
    });
}

function uploadFilePart(file, partNumber, from, size) {
    return new Promise(function(resolve, reject) {
        console.log('Uploading file part', file, from, size);

        const chunks = [];
        const rstream = fs.createReadStream(GR.sharedFolder + 'output/' + file, {
            start: from,
            end: from + size,
            autoClose: true
        })
            .on('data', function(chunk) {
                chunks.push(chunk);
            })
            .on('end', function() {
                let data = Buffer.concat(chunks).toString('binary');
                let hash = crypto.createHash('md5').update(data).digest('hex');
                checkFilePartPresence(hash)
                    .then(function(status) {
                        if (status !== false) {
                            console.log('checkFilePartPresence()', status);
                            return resolve({hash: hash, partNumber: partNumber, url: status});
                        }

                        // Upload file part
                        processUploadFilePart(hash, data)
                            .then(function(status) {
                                return resolve({hash: hash, partNumber: partNumber, url: status});
                            })
                        ;
                    })
                    .catch(error => {
                        console.log('Failed checkFilePartPresence', hash, error);
                        reject(error);
                    })
                ;
            });
    });
}

function processUploadFilePart(hash, data) {
    return new Promise(function(resolve, reject) {
        console.log('Uploading file part to the server', hash);

        // @todo figure out how to use compression, binary data is expensive when encoded in JSON

        let payload = {
            hash: hash,
            data: data,
        };
        //zlib.gzip(JSON.stringify(payload), function (error, gzippedPayload) {
            request
                .post({
                    url: GR.server + 'file-part-upload',
                    timeout: 45000,
                    body: JSON.stringify(payload),//new Buffer(gzippedPayload),
                    gzip: true,
                    //headers: {'content-type': 'application/json', 'Content-Encoding': 'gzip'}
                }, function(err, response, body) {
                    if (err || body === '') {
                        return reject(err);
                    }
                    let output = JSON.parse(body);
                    console.log('filePartUpload()', output);
                    if (output.success === false) {
                        return reject(output.error);
                    }
                    return resolve(output.url);
                })
            ;
        //});
    });
}

function checkFilePartPresence(hash) {
    return new Promise(function(resolve, reject) {
        console.log('Validating file part existance on the server', hash);

        let payload = {
            hash: hash,
        };
        request
            .post({
                url: GR.server + 'file-part-presence',
                timeout: 45000,
                body: JSON.stringify(payload),
                gzip: true,
            }, function(err, response, body) {
                if (err) {
                    return reject(err);
                }
                let output = JSON.parse(body);
                if (output.success === false) {
                    return reject(output.error);
                }
                if (output.exists === false) {
                    return resolve(false);
                }
                return resolve(output.url);
            })
        ;
    });
}

/**
 * Poll server for new input files, snapshot and termination requests
 */
function pollServer() {
    return new Promise(function(resolve, reject) {
        let payload = {
            key: GR.key,
        };
        console.log('pollServer() payload', payload);
        request
            .post({
                url: GR.server + 'poll',
                timeout: 45000,
                body: JSON.stringify(payload),
                gzip: true,
            }, function(err, response, body) {
                console.log(err);
                console.log('pollServer() body', body);
                body = JSON.parse(body);
                if (body.transitionState === 'terminate') {
                    shutdownVm(true, 'Shutdown initiated by server');
                }

                for (let stateName in body.states) {
                    if (typeof GR.states[stateName] === 'undefined') {
                        GR.states[stateName] = body.states[stateName];
                    }
                }

                console.log(GR.states);

                if (GR.activeState === false) {
                    runNextState();
                }
            })
        // @todo implement pollServer()
    });
}

function timestamp() {
    let now = new Date();
    return now.getFullYear() + '-' + (now.getMonth() + 1) + '-' + now.getDate() + ' ' + now.getHours() + ':' + now.getMinutes() + ':' + now.getSeconds();
}

function shutdownVm(success, message = undefined) {
    console.log(timestamp(), 'Shutting down VM');
    if (message) {
        console.log(message);
    }

    // @todo Send message to server about termination

    if (fs.existsSync('inhibit_shutdown')) {
        return process.exit();
    }

    exec('shutdown now', function(){});
}
